@@ -1,16 +1,15 @@
 /*
+  Copyright (c) 1990-1999 Info-ZIP.  All rights reserved.
 
- Copyright (C) 1990-1997 Mark Adler, Richard B. Wales, Jean-loup Gailly,
- Kai Uwe Rommel, Onno van der Linden and Igor Mandrichenko.
- Permission is granted to any individual or institution to use, copy, or
- redistribute this software so long as all of the original files are included,
- that it is not sold for profit, and that this copyright notice is retained.
-
+  See the accompanying file LICENSE, version 1999-Oct-05 or later
+  (the contents of which are also included in zip.h) for terms of use.
+  If, for some reason, both of these files are missing, the Info-ZIP license
+  also may be found at:  ftp://ftp.cdrom.com/pub/infozip/license.html
 */
-
 /*
  *  util.c by Mark Adler.
  */
+#define __UTIL_C
 
 #include "zip.h"
 #include "ebcdic.h"
@@ -27,7 +26,7 @@ uch upper[256], lower[256];
 #ifndef UTIL /* UTIL picks out namecmp code (all utils) */
 
 /* Local functions */
-local int recmatch OF((uch *, uch *));
+local int recmatch OF((ZCONST uch *, ZCONST uch *, int));
 local int count_args OF((char *s));
 
 #ifdef MSDOS16
@@ -56,22 +55,28 @@ char *p;                /* candidate sh expression */
 /* If p is a sh expression, a pointer to the first special character is
    returned.  Otherwise, NULL is returned. */
 {
-  for (; *p; p++)
+  for (; *p; INCSTR(p))
     if (*p == '\\' && *(p+1))
       p++;
 #ifdef VMS
     else if (*p == '%' || *p == '*')
 #else /* !VMS */
+# ifdef RISCOS
+    /* RISC OS uses # as its single-character wildcard */
+    else if (*p == '#' || *p == '*' || *p == '[')
+# else /* !RISC OS */
     else if (*p == '?' || *p == '*' || *p == '[')
+# endif
 #endif /* ?VMS */
       return p;
   return NULL;
 }
 
 
-local int recmatch(p, s)
-uch *p;       /* sh pattern to match */
-uch *s;       /* string to match it to */
+local int recmatch(p, s, cs)
+ZCONST uch *p;  /* sh pattern to match */
+ZCONST uch *s;  /* string to match it to */
+int cs;         /* flag: force case-sensitive matching */
 /* Recursively compare the sh pattern p with the string s and return 1 if
    they match, and 0 or 2 if they don't or if there is a syntax error in the
    pattern.  This routine recurses on itself no deeper than the number of
@@ -79,19 +84,27 @@ uch *s;       /* string to match it to */
 {
   unsigned int c;       /* pattern char or start of range in [-] loop */
   /* Get first character, the pattern for new recmatch calls follows */
-  c = *p++;
+  c = *POSTINCSTR(p);
 
   /* If that was the end of the pattern, match if string empty too */
   if (c == 0)
     return *s == 0;
 
-  /* '?' (or '%') matches any character (but not an empty string) */
+  /* '?' (or '%' or '#') matches any character (but not an empty string) */
 #ifdef VMS
   if (c == '%')
 #else /* !VMS */
+# ifdef RISCOS
+  if (c == '#')
+# else /* !RISC OS */
   if (c == '?')
+# endif
 #endif /* ?VMS */
-    return *s ? recmatch(p, s + 1) : 0;
+#ifdef WILD_STOP_AT_DIR
+    return (*s && *s != '/') ? recmatch(p, s + CLEN(s), cs) : 0;
+#else
+    return *s ? recmatch(p, s + CLEN(s), cs) : 0;
+#endif
 
   /* '*' matches any number of characters, including zero */
 #ifdef AMIGA
@@ -102,10 +115,18 @@ uch *s;       /* string to match it to */
   {
     if (*p == 0)
       return 1;
-    for (; *s; s++)
-      if ((c = recmatch(p, s)) != 0)
+#ifdef WILD_STOP_AT_DIR
+    for (; *s && *s != '/'; INCSTR(s))
+      if ((c = recmatch(p, s, cs)) != 0)
+        return (int)c;
+    return (*p == '/' || (*p == '\\' && p[1] == '/'))
+      ? recmatch(p, s, cs) : 2;
+#else /* !WILD_STOP_AT_DIR */
+    for (; *s; INCSTR(s))
+      if ((c = recmatch(p, s, cs)) != 0)
         return (int)c;
     return 2;           /* 2 means give up--shmatch will return false */
+#endif /* ?WILD_STOP_AT_DIR */
   }
 
 #ifndef VMS             /* No bracket matching in VMS */
@@ -113,7 +134,7 @@ uch *s;       /* string to match it to */
   if (c == '[')
   {
     int e;              /* flag true if next char to be taken literally */
-    uch *q;   /* pointer to end of [-] group */
+    ZCONST uch *q;      /* pointer to end of [-] group */
     int r;              /* flag true to match anything but the range */
 
     if (*s == 0)                        /* need a character to match */
@@ -137,55 +158,61 @@ uch *s;       /* string to match it to */
         c = *(p-1);
       else
       {
-        uch cc = case_map(*s);
+        uch cc = (cs ? *s : case_map(*s));
         if (*(p+1) != '-')
-          for (c = c ? c : *p; c <= *p; c++)    /* compare range */
-            if (case_map(c) == cc)
-              return r ? 0 : recmatch(q + 1, s + 1);
+          for (c = c ? c : (unsigned)*p; c <= (unsigned)*p; c++)
+            /* compare range */
+            if ((cs ? c : case_map(c)) == cc)
+              return r ? 0 : recmatch(q + CLEN(q), s + CLEN(s), cs);
         c = e = 0;                      /* clear range, escape flags */
       }
     }
-    return r ? recmatch(q + 1, s + 1) : 0;      /* bracket match failed */
+    return r ? recmatch(q + CLEN(q), s + CLEN(s), cs) : 0;
+                                        /* bracket match failed */
   }
 #endif /* !VMS */
 
   /* If escape ('\'), just compare next character */
   if (c == '\\')
-    if ((c = *p++) == 0)                /* if \ at end, then syntax error */
+    if ((c = *p++) == '\0')             /* if \ at end, then syntax error */
       return 0;
 
   /* Just a character--compare it */
-  return case_map(c) == case_map(*s) ? recmatch(p, ++s) : 0;
+  return (cs ? c == *s : case_map(c) == case_map(*s)) ?
+          recmatch(p, s + CLEN(s), cs) : 0;
 }
 
 
-int shmatch(p, s)
-char *p;                /* sh pattern to match */
-char *s;                /* string to match it to */
+int shmatch(p, s, cs)
+ZCONST char *p;         /* sh pattern to match */
+ZCONST char *s;         /* string to match it to */
+int cs;                 /* force case-sensitive match if TRUE */
 /* Compare the sh pattern p with the string s and return true if they match,
    false if they don't or if there is a syntax error in the pattern. */
 {
-  return recmatch((uch *) p, (uch *) s) == 1;
+  return recmatch((ZCONST uch *) p, (ZCONST uch *) s, cs) == 1;
 }
 
 
 #if defined(DOS) || defined(WIN32)
 /* XXX  also suitable for OS2?  Atari?  Human68K?  TOPS-20?? */
 
-int dosmatch(p, s)
-char *p;                /* dos pattern to match    */
-char *s;                /* string to match it to   */
+int dosmatch(p, s, cs)
+ZCONST char *p;         /* dos pattern to match    */
+ZCONST char *s;         /* string to match it to   */
+int cs;                 /* force case-sensitive match if TRUE */
 /* Treat filenames without periods as having an implicit trailing period */
 {
   char *s1;             /* revised string to match */
   int r;                /* result */
 
   if ((s1 = malloc(strlen(s) + 2)) == NULL)
-    return recmatch((uch *) p, (uch *) s) == 1;  /* will usually be correct */
+    /* will usually be OK */
+    return recmatch((ZCONST uch *) p, (ZCONST uch *) s, cs) == 1;
   strcpy(s1, s);
   if (strchr(p, '.') && !strchr(s1, '.'))
     strcat(s1, ".");
-  r = recmatch((uch *)p, (uch *)s1);
+  r = recmatch((ZCONST uch *)p, (ZCONST uch *)s1, cs);
   free((zvoid *)s1);
   return r == 1;
 }
@@ -193,8 +220,8 @@ char *s;                /* string to match it to   */
 #endif /* DOS || WIN32 */
 
 zvoid far **search(b, a, n, cmp)
-zvoid *b;               /* pointer to value to search for */
-zvoid far **a;          /* table of pointers to values, sorted */
+ZCONST zvoid *b;        /* pointer to value to search for */
+ZCONST zvoid far **a;   /* table of pointers to values, sorted */
 extent n;               /* number of pointers in a[] */
 int (*cmp) OF((ZCONST zvoid *, ZCONST zvoid far *)); /* comparison function */
 
@@ -206,15 +233,15 @@ int (*cmp) OF((ZCONST zvoid *, ZCONST zvoid far *)); /* comparison function */
    left up to the cmp() function.  A binary search is used, and it is
    assumed that the list is sorted in ascending order. */
 {
-  zvoid far **i;        /* pointer to midpoint of current range */
-  zvoid far **l;        /* pointer to lower end of current range */
+  ZCONST zvoid far **i; /* pointer to midpoint of current range */
+  ZCONST zvoid far **l; /* pointer to lower end of current range */
   int r;                /* result of (*cmp)() call */
-  zvoid far **u;        /* pointer to upper end of current range */
+  ZCONST zvoid far **u; /* pointer to upper end of current range */
 
-  l = (zvoid far **)a;  u = l + (n-1);
+  l = (ZCONST zvoid far **)a;  u = l + (n-1);
   while (u >= l) {
     i = l + ((unsigned)(u - l) >> 1);
-    if ((r = (*cmp)(b, (char *)*(struct zlist **)i)) < 0)
+    if ((r = (*cmp)(b, (ZCONST char far *)*(struct zlist far **)i)) < 0)
       u = i - 1;
     else if (r > 0)
       l = i + 1;
@@ -244,7 +271,7 @@ void init_upper()
   struct country far *info = &country_info;
   union REGS regs;
   struct SREGS sregs;
-  int c;
+  unsigned int c;
 
   regs.x.ax = 0x3800; /* get country info */
   regs.x.dx = FP_OFF(info);
@@ -274,7 +301,7 @@ void init_upper()
 
 void init_upper()
 {
-  int c;
+  unsigned int c;
 #if defined(ATARI) || defined(CMS_MVS)
 #include <ctype.h>
 /* this should be valid for all other platforms too.   (HD 11/11/95) */
@@ -283,9 +310,9 @@ void init_upper()
     lower[c] = isupper(c) ? tolower(c) : c;
   }
 #else
-  for (c = 0; c < sizeof(upper); c++) upper[c] = lower[c] = c;
-  for (c = 'a'; c <= 'z';        c++) upper[c] = c - 'a' + 'A';
-  for (c = 'A'; c <= 'Z';        c++) lower[c] = c - 'A' + 'a';
+  for (c = 0; c < sizeof(upper); c++) upper[c] = lower[c] = (uch)c;
+  for (c = 'a'; c <= 'z';        c++) upper[c] = (uch)(c - 'a' + 'A');
+  for (c = 'A'; c <= 'Z';        c++) lower[c] = (uch)(c - 'A' + 'a');
 #endif
 }
 #  endif /* !OS2 */
@@ -293,7 +320,7 @@ void init_upper()
 #endif /* ?MSDOS16 */
 
 int namecmp(string1, string2)
-  char *string1, *string2;
+  ZCONST char *string1, *string2;
 /* Compare the two strings ignoring case, and correctly taking into
  * account national language characters. For operating systems with
  * case sensitive file names, this function is equivalent to strcmp.
@@ -373,9 +400,55 @@ char *str_oem_to_iso(dst, src)
 #endif
 
 
+
+/* DBCS support for Info-ZIP's zip  (mainly for japanese (-: )
+ * by Yoshioka Tsuneo (QWF00133@nifty.ne.jp,tsuneo-y@is.aist-nara.ac.jp)
+ * This code is public domain!   Date: 1998/12/20
+ */
+#ifdef _MBCS
+
+char *___tmp_ptr;
+
+int lastchar(ptr)
+    ZCONST char *ptr;
+{
+    ZCONST char *oldptr = ptr;
+    while(*ptr != '\0'){
+        oldptr = ptr;
+        INCSTR(ptr);
+    }
+    return (int)(unsigned)*oldptr;
+}
+
+unsigned char *zmbschr(str, c)
+    ZCONST unsigned char *str;
+    unsigned int c;
+{
+    while(*str != '\0'){
+        if (*str == c) {return (char*)str;}
+        INCSTR(str);
+    }
+    return NULL;
+}
+
+unsigned char *zmbsrchr(str, c)
+    ZCONST unsigned char *str;
+    unsigned int c;
+{
+    unsigned char *match = NULL;
+    while(*str != '\0'){
+        if (*str == c) {match = (char*)str;}
+        INCSTR(str);
+    }
+    return match;
+}
+#endif /* _MBCS */
+
+
+
 #ifndef UTIL
 
-extern char *getenv OF((ZCONST char *));
+extern char *getenv();
 
 /*****************************************************************
  | envargs - add default options from environment to command line
@@ -405,11 +478,11 @@ void envargs(Pargc, Pargv, envstr, envstr2)
     /* see if anything in the environment */
     envptr = getenv(envstr);
     if (envptr != NULL)                                /* usual var */
-        while (isspace(*envptr))           /* we must discard leading spaces */
+        while (isspace((uch)*envptr))      /* we must discard leading spaces */
             envptr++;
     if (envptr == NULL || *envptr == '\0')
         if ((envptr = getenv(envstr2)) != NULL)                 /* alternate */
-            while (isspace(*envptr))
+            while (isspace((uch)*envptr))
                 envptr++;
     if (envptr == NULL || *envptr == '\0')
         return;
@@ -438,20 +511,22 @@ void envargs(Pargc, Pargv, envstr, envstr2)
         if (*bufptr == '"') {
             char *argstart = ++bufptr;
             *(argv++) = argstart;
-            for (ch = *bufptr; ch != '\0' && ch != '\"'; ch = *(++bufptr))
+            for (ch = *bufptr; ch != '\0' && ch != '\"';
+                 ch = *PREINCSTR(bufptr))
                 if (ch == '\\' && bufptr[1] != '\0')
-                    ++bufptr;                  /* skip char after backslash */
-            if (ch != '\0') *(bufptr++) = '\0';     /* overwrite trailing " */
+                    ++bufptr;               /* skip to char after backslash */
+            if (ch != '\0')                       /* overwrite trailing '"' */
+                *(bufptr++) = '\0';
 
             /* remove escape characters */
-            while (argstart = strchr(argstart, '\\')) {
+            while ((argstart = MBSCHR(argstart, '\\')) != NULL) {
                 strcpy(argstart, argstart + 1);
                 if (*argstart)
                     ++argstart;
             }
         } else {
             *(argv++) = bufptr;
-            while ((ch = *bufptr) != '\0' && !isspace(ch)) ++bufptr;
+            while ((ch = *bufptr) != '\0' && !isspace((uch)ch)) INCSTR(bufptr);
             if (ch != '\0') *(bufptr++) = '\0';
         }
 #else
@@ -461,20 +536,20 @@ void envargs(Pargc, Pargv, envstr, envstr2)
         /* separators and double quotes are illegal in filenames.  */
         if (*bufptr == '"') {
             *(argv++) = ++bufptr;
-            while ((ch = *bufptr) != '\0' && ch != '\"') ++bufptr;
+            while ((ch = *bufptr) != '\0' && ch != '\"') INCSTR(bufptr);
             if (ch != '\0') *(bufptr++) = '\0';
         } else {
             *(argv++) = bufptr;
-            while ((ch = *bufptr) != '\0' && !isspace(ch)) ++bufptr;
+            while ((ch = *bufptr) != '\0' && !isspace((uch)ch)) INCSTR(bufptr);
             if (ch != '\0') *(bufptr++) = '\0';
         }
 #  else
         *(argv++) = bufptr;
-        while ((ch = *bufptr) != '\0' && !isspace(ch)) ++bufptr;
+        while ((ch = *bufptr) != '\0' && !isspace((uch)ch)) INCSTR(bufptr);
         if (ch != '\0') *(bufptr++) = '\0';
 #  endif
 #endif /* ?(AMIGA || UNIX) */
-        while ((ch = *bufptr) != '\0' && isspace(ch)) ++bufptr;
+        while ((ch = *bufptr) != '\0' && isspace((uch)ch)) INCSTR(bufptr);
     } while (ch);
 
     /* now save old argc and copy in the old args */
@@ -500,25 +575,26 @@ char *s;
         ++count;
 #if defined(AMIGA) || defined(UNIX)
         if (*s == '\"') {
-            for (ch = *(++s); ch != '\0' && ch != '\"'; ch = *(++s))
+            for (ch = *PREINCSTR(s); ch != '\0' && ch != '\"';
+                 ch = *PREINCSTR(s))
                 if (ch == '\\' && s[1] != '\0')
-                    ++s;
-            if (*s) ++s;        /* trailing quote */
+                    INCSTR(s);
+            if (*s) INCSTR(s);  /* trailing quote */
         } else
-            while ((ch = *s) != '\0' && !isspace(ch)) ++s;
+            while ((ch = *s) != '\0' && !isspace((uch)ch)) INCSTR(s);
 #else
 #  ifdef WIN32
         if (*s == '\"') {
             ++s;                /* leading quote */
-            while ((ch = *s) != '\0' && ch != '\"') ++s;
-            if (*s) ++s;        /* trailing quote */
+            while ((ch = *s) != '\0' && ch != '\"') INCSTR(s);
+            if (*s) INCSTR(s);  /* trailing quote */
         } else
-            while ((ch = *s) != '\0' && !isspace(ch)) ++s;
+            while ((ch = *s) != '\0' && !isspace((uch)ch)) INCSTR(s);
 #  else
-        while ((ch = *s) != '\0' && !isspace(ch)) ++s;
+        while ((ch = *s) != '\0' && !isspace((uch)ch)) INCSTR(s);
 #  endif
 #endif /* ?(AMIGA || UNIX) */
-        while ((ch = *s) != '\0' && isspace(ch)) ++s;
+        while ((ch = *s) != '\0' && isspace((uch)ch)) INCSTR(s);
     } while (ch);
 
     return(count);
@@ -586,7 +662,7 @@ void expand_args(argcp, argvp)
       /* substitute new argument list in place of old one */
       *argcp = newargc;
       *argvp = newargv;
-#else /* ?DOS */
+#else /* !DOS */
       if (argcp || argvp) return;
 #endif /* ?DOS */
 }
@@ -617,4 +693,4 @@ int printnames()
      return 0;
 }
 
-#endif
+#endif /* DEBUGNAMES */
